#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)


package Deconv;
use strict;
use warnings;
use PDL;
use Readonly;
use Storable;
use File::Temp qw (tempfile);
use SGE_API;
use Submit_tools;
use Carp;
use Cwd 'abs_path';
use File::Basename;
use Hash::Util qw (lock_keys);
use Archive::Tar;
use FindBin qw($Bin);
use Data::Dumper;


Readonly my $SPACE                  => q{ };
Readonly my $EPSILON                => '0.000000001';
Readonly my $LOCATION_OF_MAN_SCRIPTS => "$Bin/Cluster/Remote_Scripts";
Readonly my $REMOTE_SCRIPT_NANE => 'trainer.pl';
Readonly my @MANDATORY_SCRIPTS   => qw (Model3.pm
                                        Model.pm
                                        trainer.pl
                                        svd);

Readonly my $D_UNDEFINED_VALUE_IN_MATRIX => '25';

Readonly my $RANDOM_WALK_SESSION =>'1';
Readonly my $SVM_SESSION         =>'2';
Readonly my $CYCLE_SESSION       =>'3';


Readonly my $MEMORY_FOR_CYCLE_SESSION =>'8G'; # NOTICE former '4G'

#test();

sub test {

    my @p1 = (0.965,0.1,1);
    my @p2 = (0.975,0.1,1);



    my $deconv = Deconv->new;
    $deconv->verbose;

#    $deconv->set_cycle_train_range(1,1,50);
    $deconv->set_train_cycles(50);
    $deconv->set_number_of_cycles(50);
    $deconv->set_counter_threshold(200);
    $deconv->set_c_g(1,1);
    $deconv->set_random_walk_range(\@p1,\@p2);
    $deconv->set_file_pos_to_seq('../tmp/s_8_00001.hash');#s_8_00001.hash#pos_to_seq_8_0001.hash
    $deconv->set_unify_int_file('../tmp/s_8_0001_int.txt');
    $deconv->test_verify_loading_params;
    $deconv->set_cross_val(2);
    #$deconv->set_bring_back_files_to('../test/');

      $deconv->test_preparation_all;
      $deconv->generate_command_line;
      $deconv->debug_print_parameters;
      $deconv->flag_debug_mode_set;
      $deconv->debug_print_command_line;
      $deconv->qsub_all;
#         $deconv->debug_print_qsub_script;
      $deconv->submit_life_time_set('7h');
      $deconv->submit_quicky(1);
      $deconv->qsub_bye;

       $deconv->stdout_parse_matrix_quickly;
       $deconv->debug_print_all_matrices;
       $deconv->debug_print_maximum_for_each_martix;
       $deconv->matanl_maximum_upon_average;
}




sub new {
        #Create a new object
	my $classname = shift;
	my $self = {};
	bless $self,$classname;
        $self->add_mandatory_files;
	return $self;
}

sub DESTROY {
    my $self = shift;
    if ($self->flag_debug_mode_get) {
        printf ("$self dying at %s\n",scalar localtime);
    }
    return 1;
}



sub add_mandatory_files {
    my $self = shift;

    my $folder = $LOCATION_OF_MAN_SCRIPTS;

    foreach my $script_name (@MANDATORY_SCRIPTS) {

        my $filename = abs_path ("$folder/$script_name");
        confess "can't find mandatory file $filename"
            if (not -e $filename);

        my $key      = fileparse ($filename,qr/\.[^.]*/);
        $self->{mandatory_files}->{$key}->{name} = $filename;
    }

    $self->flag_all_mandatory_files_verified_set;
    return 1;
}

sub get_mandatory_files {
    my $self = shift;

    my @filenames;
    my @file_keys = keys (%{$self->{mandatory_files}});
    foreach my $key (@file_keys) {
        push @filenames, $self->{mandatory_files}->{$key}->{name};
    }
    return (\@file_keys,\@filenames);
}


################################### tests #############################
sub test_preparation_all {
    my $self = shift;
    $self->test_all_required_data_files;
    $self->test_verify_loading_params;
    $self->test_mandatory_files;
    return 1;
}

sub test_mandatory_files {
    my $self = shift;


    if ($self->flag_verbose_get) {
        print "Test all mandatory files:\n";
        my ($r_keys,$r_names) = $self->get_mandatory_files;
        print "\t\t\t$_\n" foreach (@$r_names);
    }


    if ($self->flag_all_mandatory_files_verified_get) {
        if ($self->flag_verbose_get) {
            print "\t\t\t...ok\n";
        }
        return 1;
    }
    confess "test mandatory files failed\n";
    return;
}

sub test_all_required_data_files {

    my $self = shift;
    my @test = qw (
                    test_file_pos_to_seq
                    test_unify_int_file
                  );
    foreach my $test (@test) {
        $self->$test;
    }
    $self->flag_all_files_verified_set;
    return 1;
}

sub test_verify_loading_params {

     my $self = shift;

     if ($self->flag_verbose_get) {
         print "Test all requiered parameters loaded:\t";
     }

     if (
         ($self->get_file_pos_to_seq)                                        and
         ($self->get_number_of_cycles)                                       and
         ($self->get_unify_int_file)                                         and
         ($self->get_counter_threshold)                                      and

           (
            (
             ($self->flag_session__are_we_in_random_walk_session)and
             ($self->get_c_g)and
             ($self->get_train_cycles)
            )
                                                                      or
            (
             ($self->flag_session__are_we_in_svm_session)and
             ($self->get_p1_p2)and
             ($self->get_train_cycles)
            )
                                                                      or
            (
             ($self->flag_session__are_we_in_cycle_session)and
             ($self->get_c_g)and
             ($self->get_p1_p2)
            )
        )
      )                                        {

        if ($self->flag_verbose_get) {print "..ok\n"};
        $self->flag_all_params_loaded_set;
        return 1;
    }
    $self->debug_print_parameters;
    confess "not all params loaded\n";
}


sub test_file_pos_to_seq {

    my $self = shift;
    my $file = $self->get_file_pos_to_seq;

    if ($self->flag_verbose_get) {
        print "Test file pos_to_seq(and trying to open it):\t";
    }
    if ((not $file) or (not -e $file)) {
        confess "can't find file_pos_to_seq: $file\n";
        return;
    }
    my $del = retrieve($file);
    if ($self->flag_verbose_get) {print "..ok\n"}
    return 1;
}

sub test_unify_int_file {
    my $self = shift;
    my $file = $self->get_unify_int_file;
    if ($self->flag_verbose_get) {
        print "Test unified intensity file:\t";
    }
    if ((not $file) or  (not -e $file)) {
        confess "can't find unified intensity file: $file\n";
        return;
    }
    if ($self->flag_verbose_get) {print "..ok\n"};
    return 1;
}

sub test_location_stdout_files {
    my $self = shift;
    my @files = $self->stdout_files_get;

    if ($self->flag_verbose_get) {
        print "Test stdout files", 
              "(the files that return from the nodes):\t";
    }

    if (not @files) {
        confess "there are no files to check\n";
    }
    foreach my $file (@files) {
        confess "can't find $file\n"
        if (not -e $file)
    }
    if ($self->flag_verbose_get) {
        print "..ok ","found ",scalar @files," files\n";
    }
    return 1;
}

sub test_stdout_do_we_have_all_files {

    my $self = shift;
    my $n_files = scalar $self->stdout_files_get;
    my $cluster_units = $self->get_range_in_cluster_units;

    if ($self->flag_verbose_get) {
        print "Test number of files returned from the cluster:\t";
    }

    if ($n_files < $cluster_units) {
        carp "can't find all report files from the cluster\n",
             "Maybe life time issue?\t", "I will try to continue processing\n";
    }
    elsif ($n_files == $cluster_units) {
        if ($self->flag_verbose_get) {
            print "..ok\n";
        }
        return 1;
    }
    else {
        confess "Identified more files than possible: $n_files", 
              "(expecting $cluster_units)\n";
    }
}



sub verbose {
    my $self = shift;
    $self->flag_verbose_set;
    return 1;
}



sub test_returen_tar {

    my $self = shift;
    my $filename = shift ||
        confess "must get filename\n";
    my $r_lookup_table_for_file = shift;
    confess "test_returen_tar: second argument must be ref to hash\n"
        if (ref $r_lookup_table_for_file ne 'HASH');


    if ($self->flag_verbose_get) {
        print "Test returned tar - $filename for the cluster\t";
    }


    if (not -e $filename) {
        confess "can't find filename\n";
    }

    if ($self->flag_verbose_get) {
        print "..Found file\t";
    }

    my $tar = Archive::Tar->new();
    $tar->read($filename,1);
    my @files_in_tar = $tar->list_files;
    my @token = grep {m/token/} @files_in_tar;
    confess "problem to find to token in $filename.",
          " List of files is @files_in_tar\n"
        if (scalar @token != 1);


    if ($self->flag_verbose_get) {
        print "..Opened and found token\t";
    }

    my $token_content = $tar->get_content(@token);
    my @token_lines = split /\n+/, $token_content;
    foreach my $line (@token_lines) {
        next if (not ($line =~m/^>(\w+):(.+)/));
        my $k = $1;
        my $v = $2;
        if (not $r_lookup_table_for_file->{$k} eq $v) {
            confess "Problem reading token in file $filename.",
                    " $v is not the value of $k. It should be:",
                    $r_lookup_table_for_file->{$k},"\n";
        }
    }

    if ($self->flag_verbose_get) {
        print "..ok (token passed the test)\t";
    }
    return 1;
}


######################################## flags #####################
sub flag_all_files_verified_set {
    my $self = shift;
    $self->{meta}->{flags}->{all_files} = 1;
    return $self->{meta}->{flags}->{all_files};
}

sub flag_all_files_verified_get {
    my $self = shift;
    return $self->{meta}->{flags}->{all_files};
}

sub flag_all_mandatory_files_verified_set {
    my $self = shift;
    $self->{meta}->{flags}->{mandatory} = 1;
    return $self->{meta}->{flags}->{mandatory};
}

sub flag_all_mandatory_files_verified_get {
    my $self = shift;
    return $self->{meta}->{flags}->{mandatory};
}

sub flag_all_params_loaded_set {
    my $self = shift;
    $self->{meta}->{flags}->{all_param} = 1;
    return $self->{meta}->{flags}->{all_param};
}

sub flag_all_params_loaded_get {
    my $self = shift;
    return $self->{meta}->{flags}->{all_param}
}






################## Session type ################
sub flag_random_walk_session_set {
    my $self = shift;
    $self->flag_session_type_set($RANDOM_WALK_SESSION);
    $self->{meta}->{flags}->{random_walk_session} = 1;
    return 1;
}

sub flag_svm_param_session_set {
    my $self = shift;
    $self->flag_session_type_set($SVM_SESSION);
    $self->{meta}->{flags}->{svm_params_session} = 1;
    return 1;
}
sub flag_cycle_range_session_set {
    my $self = shift;
    $self->flag_session_type_set($CYCLE_SESSION);
    $self->{meta}->{flags}->{cycle_session} = 1;
    return 1;
}
#
sub flag_session_type_set {
    my $self = shift;

    my $type = shift ||
        confess "flag_session_type_set error - must give session type\n";

    if ($self->flag_session_type_get) {
        confess "can't define a new session\n";
    }

    $self->{meta}->{flags}->{session_type} = $type;
    return $self->{meta}->{flags}->{session_type};
}
sub flag_session_type_get {
    my $self = shift;
    return $self->{meta}->{flags}->{session_type};
}
#

sub flag_session__are_we_in_random_walk_session {
    my $self = shift;
    return $self->{meta}->{flags}->{random_walk_session};
}

sub flag_session__are_we_in_svm_session {
    my $self = shift;
    return $self->{meta}->{flags}->{svm_params_session}
}

sub flag_session__are_we_in_cycle_session {
    my $self = shift;
    return $self->{meta}->{flags}->{cycle_session}
}

#########################################################


sub flag_svm_range_set {
    my $self = shift;
    $self->{meta}->{flags}->{svm_range_set} = 1;
    return $self->{meta}->{flags}->{svm_range_set};
}
sub flag_svm_range_get {
    my $self = shift;
    return $self->{meta}->{flags}->{svm_range_set};
}
sub flag_random_walk_range_set {
    my $self = shift;
    $self->{meta}->{flags}->{random_walk_range_set} = 1;
    return $self->{meta}->{flags}->{random_walk_range_set};
}
sub flag_random_walk_range_get {
    my $self = shift;
    return $self->{meta}->{flags}->{random_walk_range_set};
}




sub flag_command_line_set {
    my $self = shift;
    $self->{meta}->{flags}->{command_line} = 1;
    return 1;
}
sub flag_command_line_get {
    my $self = shift;
    return $self->{meta}->{flags}->{command_line};
}
sub flag_verbose_set {
    my $self = shift;
    $self->{meta}->{flags}->{verbose}=1;
    return $self->{meta}->{flags}->{verbose};
}
sub flag_verbose_get {
    my $self = shift;
    return $self->{meta}->{flags}->{verbose};
}
sub flag_qsub_range_set {
    my $self = shift;
    $self->{meta}->{flags}->{qsub_range} = 1;
    return $self->{meta}->{flags}->{qsub_range};
}

sub flag_qsub_range_get {
    my $self = shift;
    return $self->{meta}->{flags}->{qsub_range};
}
sub flag_debug_mode_set {
    my $self = shift;
    $self->{meta}->{flags}->{debug_mode} = 1;
    return $self->{meta}->{flags}->{debug_mode};
}
sub flag_debug_mode_get {
    my $self = shift;
    return $self->{meta}->{flags}->{debug_mode};
}

sub flag_finish_submitter_set {
    my $self = shift;
    $self->{meta}->{flags}->{finish_submitter} = 1;
    return $self->{meta}->{flags}->{finish_submitter};
}

sub flag_finish_submitter_get {
    my $self = shift;
    return $self->{meta}->{flags}->{finish_submitter};
}

sub flag_SVM_files_generation_set {
    my $self = shift;
    confess "can't set files generation while training (cross validation) is set\n"
        if ($self->flag_SVM_training_get);

    $self->{meta}->{flags}->{svm_file_gen} = 1;
    return $self->{meta}->{flags}->{svm_file_gen};
}

sub flag_SVM_files_generation_get {
    my $self = shift;
    return $self->{meta}->{flags}->{svm_file_gen};
}
sub flag_SVM_training_set {
    my $self = shift;
    confess "can't set training (cross validation) while files generation is set\n"
        if ($self->flag_SVM_files_generation_get);
    $self->{meta}->{flags}->{svm_train} = 1;
}
sub flag_SVM_training_get {
    my $self = shift;
    return $self->{meta}->{flags}->{svm_train};
}



######################################### parameters #########################
sub set_random_walk_range {
    my ($self,$r_p1_range, $r_p2_range) = @_;



    confess "set_random_walk_range error\n"
        if (
            (scalar @$r_p1_range != 3)          or
            (scalar @$r_p2_range != 3)          or
            (maxI(@$r_p1_range)  >  1)          or
            (maxI(@$r_p2_range)  >  1)
                                       );

    $self->{config}->{p1}->{start} = $r_p1_range->[0];
    $self->{config}->{p1}->{gap}   = $r_p1_range->[1];
    $self->{config}->{p1}->{stop}  = $r_p1_range->[2];

    $self->{config}->{p2}->{start} = $r_p2_range->[0];
    $self->{config}->{p2}->{gap}   = $r_p2_range->[1];
    $self->{config}->{p2}->{stop}  = $r_p2_range->[2];

    $self->flag_random_walk_session_set;
    $self->flag_random_walk_range_set;

    return 1;
}

sub get_random_walk_range {
    my $self = shift;


    return 
        if (not $self->flag_session__are_we_in_random_walk_session);

    my @p1 = (
                $self->{config}->{p1}->{start},
                $self->{config}->{p1}->{gap},
                $self->{config}->{p1}->{stop}
               );
    my @p2 = (
                $self->{config}->{p2}->{start},
                $self->{config}->{p2}->{gap},
                $self->{config}->{p2}->{stop}
               );

    return (\@p1,\@p2);
}

sub set_svm_range {
    my $self = shift;

    my ($r_c_range, $r_g_range) = @_;
    confess "set_svm_range error\n"
        if (
            (scalar @$r_c_range != 3)  or
            (scalar @$r_g_range != 3)
                                       );
    $self->{config}->{c}->{start} = $r_c_range->[0];
    $self->{config}->{c}->{gap}   = $r_c_range->[1];
    $self->{config}->{c}->{stop}  = $r_c_range->[2];

    $self->{config}->{g}->{start} = $r_g_range->[0];
    $self->{config}->{g}->{gap}   = $r_g_range->[1];
    $self->{config}->{g}->{stop}  = $r_g_range->[2];

    $self->flag_svm_param_session_set;
    $self->flag_svm_range_set;

    return 1;
}

sub get_svm_range {
    my $self = shift;

    return 
        if (not $self->flag_session__are_we_in_svm_session);

    my @c = (
                $self->{config}->{c}->{start},
                $self->{config}->{c}->{gap}  ,
                $self->{config}->{c}->{stop} ,
               );

    my @g = (
                $self->{config}->{g}->{start},
                $self->{config}->{g}->{gap}  ,
                $self->{config}->{g}->{stop} ,
               );
    return (\@c,\@g);
}

sub set_p1_p2 {
    my ($self,$p1,$p2) = @_;

    confess "set_p1_p2 error\n"
        if (
            (not defined $p1)       or 
            (not defined $p2)       or
            ($p1 > 1)               or
            ($p2 > 1)               or
            ($p1 < 0)               or
            ($p2 < 0)
           );

    $self->{config}->{p1} = $p1;
    $self->{config}->{p2} = $p2;

    return ($self->{config}->{p1},
            $self->{config}->{p2});
}


sub get_p1_p2 {

    my ($self) = @_;
    return
        if ($self->flag_session__are_we_in_random_walk_session);

    return ($self->{config}->{p1},
            $self->{config}->{p2});
}

sub set_c_g {

    my ($self,$c,$g) = @_;

    confess "set_c_g error\n"
        if (not defined $c);

    $self->{config}->{c} = $c;
    $self->{config}->{g} = $g;

    return ($self->{config}->{c},
            $self->{config}->{g});

}

sub get_c_g {
    my $self = shift;
    return 
        if ($self->flag_session__are_we_in_svm_session);

    return ($self->{config}->{c},
            $self->{config}->{g});
}

######################################### Train Cycle session #########################
sub set_cycle_train_range {
    my $self = shift;
    confess "set_cycle_range error (ecpecting array of 3 parameters)\n"
        if (scalar @_< 3);

    my ($start,$stop,$gap) = @_;
    if (
        (int $start != $start)  or
        (int $stop  != $stop )  or
        (int $gap   != $gap  )  or
        ($start<1)            ){
        confess "set_cycle_range input error\n";
    }

    $self->flag_cycle_range_session_set;

    $self->{config}->{train_cycle_range}->{start} = $start;
    $self->{config}->{train_cycle_range}->{stop}  = $stop;
    $self->{config}->{train_cycle_range}->{gap}   = $gap;

    return 1;
}

sub get_cycle_train_range {
    my $self = shift;

    return if (not $self->flag_session__are_we_in_cycle_session);

    my @range = ($self->{config}->{train_cycle_range}->{start},
                 $self->{config}->{train_cycle_range}->{stop},
                 $self->{config}->{train_cycle_range}->{gap});

    #the range return in array format like c_g and p1_p2
    return \@range;
}


sub set_train_cycles {
    my $self = shift;
    if (@_) {
        @{$self->{config}->{train_cycles}} = @_;
        return @{$self->{config}->{train_cycles}};
    }
    return;
}
sub get_train_cycles {
    my $self = shift;
    return if ($self->flag_session__are_we_in_cycle_session);
    return @{$self->{config}->{train_cycles}};
}
######################################### more parameters #########################
sub set_file_pos_to_seq {
    my $self = shift;
    if (@_) {
        my $p = shift;
        $self->{config}->{file_pos_to_seq} = 
            $p;
        return $self->{config}->{file_pos_to_seq};
    }
    return;
}

sub get_file_pos_to_seq {
    my $self = shift;
    return $self->{config}->{file_pos_to_seq};
}

sub set_number_of_cycles {
    my $self = shift;
    if (@_) {
        my $p = shift;
        $self->{config}->{number_of_cycles} = 
            $p;
        return $self->{config}->{number_of_cycles};
    }
    return;
}

sub get_number_of_cycles {
    my $self = shift;
    return $self->{config}->{number_of_cycles}
}

sub set_unify_int_file {
    my $self = shift;
    if (@_) {
        my $p = shift;
        $self->{config}->{unify_int_file} =
            $p;
        return $self->{config}->{unify_int_file};
    }
    return;

}
sub get_unify_int_file {
    my $self = shift;
    return $self->{config}->{unify_int_file};
}
sub set_counter_threshold {
    my $self = shift;
    if (@_) {
        my $p = shift;
        $self->{config}->{counter_threshold} = 
            $p;
        return $self->{config}->{counter_threshold};
    }
    return;

}
sub get_counter_threshold {
    my $self = shift;
    return $self->{config}->{counter_threshold};
}

sub set_estimate_prop {
    my $self = shift;
    $self->{config}->{estimate_prop} = 1;
    return $self->{config}->{estimate_prop}
}
sub get_estimate_prop {
    my $self = shift;
    return $self->{config}->{estimate_prop};
}

sub set_cross_val {
    my $self = shift;

    $self->flag_SVM_training_set;
    if (@_) {
        my $p = shift;
        $self->{config}->{cross_val} = $p;
        return $self->{config}->{cross_val};
    }
    return;
}
sub get_cross_val {
    my $self = shift;
    return $self->{config}->{cross_val};
}

sub set_bring_back_files_to {
    my $self = shift;
    my $dir = shift ||
        confess "must give folder name\n";
    $self->flag_SVM_files_generation_set;


    if (not -d $dir) {
        mkdir $dir ||
            confess "can't make dir $dir:$!\n";
    }
    $dir = abs_path $dir;
    $self->{config}->{get_files_to} = $dir;
    return $self->{config}->{get_files_to};
}
sub get_bring_back_files_to {
    my $self = shift;
    return $self->{config}->{get_files_to};
}
######################################### Debug #########################

sub debug_print_stdout_filename {
    my $self = shift;
    my @filenames = $self->stdout_files_get;
    print "\n",
    "--------------------------------------------------\n",
    "@filenames ",
    "\n",
    "--------------------------------------------------\n";
    return 1;
}



sub debug_print_command_line {
    my $self = shift;
    print "\n",
    "--------------------------------------------------\n",
    $self->{command_line},"\n",
    "--------------------------------------------------\n";
    print "\n";
    return 1;
}

sub debug_print_parameters {
    my $self = shift;

    my $svm_params;
    my $random_walk;
    my $train_cycles;
    my $hash_pos_to_seq_file    = $self->get_file_pos_to_seq                || 'undef';
    my $number_of_cycles        = $self->get_number_of_cycles               || 'undef';
    my $unify_int_file          = $self->get_unify_int_file                 || 'undef';
    my $counter_threshold       = $self->get_counter_threshold              || 'undef';
    my $cross_vald              = $self->get_cross_val                      || 'undef';
    my $estimate_prop           = $self->get_estimate_prop                  || 'undef';
    my $bring_files             = $self->get_bring_back_files_to            || 'undef';




    if ($self->flag_session__are_we_in_random_walk_session) {
        $random_walk            = 'Wild card';
        $svm_params             = join $SPACE, $self->get_c_g;
        $train_cycles           = join $SPACE, $self->get_train_cycles;
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        $random_walk            = join $SPACE, $self->get_p1_p2;
        $svm_params             = 'Wild card';
        $train_cycles           = join $SPACE, $self->get_train_cycles;
    }
    elsif ($self->flag_session__are_we_in_cycle_session) {
        $random_walk            = join $SPACE, $self->get_p1_p2;
        $svm_params             = join $SPACE, $self->get_c_g;
        $train_cycles           = 'Wild card';
    }



    print "\n",
          "--------------------------------------------------\n",
          "Parameters:\n",
          "Hash_pos_to_seq_file:\t", $hash_pos_to_seq_file  ,"\n",
          "Number_of_cycles:\t",     $number_of_cycles      ,"\n",
          "Unify_int_file:\t",       $unify_int_file        ,"\n",
          "Counter_threshold:\t",    $counter_threshold     ,"\n",
          "Cross_vald:\t",           $cross_vald            ,"\n",
          "Estimate_prop:\t",        $estimate_prop         ,"\n",
          "Bring file to:\t",        $bring_files           ,"\n",
          "Train_cycles:\t",         $train_cycles          ,"\n",
          "SVM_params:\t",           $svm_params            ,"\n",
          "Random_walk\t",           $random_walk           ,"\n",
          "--------------------------------------------------\n",
          "\n";
}

sub debug_print_qsub_script {
    my $self = shift;
    my $filename = $self->get_sge_script;
    print "\n",
          "--------------------------------------------------\n",
    open my $FH,'<', $filename
        or confess "can't open $filename: $!\n";

    while (<$FH>) {
        print $_;
    }
    print "--------------------------------------------------\n",
          "\n";
    close $FH;
    return 1;
}

sub debug_print_matrix {
    my $self = shift;
    my $cycle = shift ||
        confess "debug_print_matrix error: ",
              "must enter cycle\n";


    my $matrix = $self->stdout_matrix_get($cycle); 
    confess "debug_print_matrix error: ",
          "matrix is not defined for cycle: $cycle\n"
            if (ref $matrix ne 'PDL');

    print "\n",
          "--------------------------------------------------\n",
          "cycle:$cycle\n",
           $matrix,
          "--------------------------------------------------\n",
          "\n";
    return 1;
}

sub debug_print_all_matrices {
    my $self = shift;
    my @cycles = $self->stdout_matrix_get_cycles_with_matrix;
    foreach my $cycle (@cycles) {
        $self->debug_print_matrix($cycle);
    }
    return 1;
}

sub debug_print_maximum_for_each_martix {
    my $self = shift;
    my @cycles = $self->stdout_matrix_get_cycles_with_matrix;

    print "\n",
          "--------------------------------------------------\n",
          "Finding maximum points for each cycle:\n";

    foreach my $cycle (@cycles) {
        my $matrix = $self->stdout_matrix_get($cycle);
        my ($par1_max,$par2_max,$max) = $self->matanl_find_maximum_point($matrix);
        print "Cycle: $cycle - maximum is $max in $par1_max and $par2_max\n";
    }
    return 1;
}



######################################### qsub #########################
sub qsub_all {
    #you should goodbye qsub ($sge->bye) later!
    my $self = shift;
    $self->test_all_required_data_files
        if (not flag_all_files_verified_get);
    $self->generate_command_line;
    $self->qsub_script_set;
    $self->qsub_setting_range;
    $self->qsub_script_add_all_files;
    $self->qsub_add_command_line;
    my $filename = 
        $self->qsub_compile_and_test;

    return $filename;
}

sub qsub_bye {
    my $self = shift;
    my $sge = $self->qsub_script_get;
    $sge->bye;
    return 1;
}

sub qsub_script_set {
    my $self = shift;
    my $sge = SGE_API->new();
    $self->{sge} = $sge;
    return $sge;
}

sub qsub_script_get {
    my $self = shift;
    return $self->{sge};
}

sub qsub_setting_range {
    my $self = shift;

    my $sge = $self->qsub_script_get;

    my @starts;
    my @gaps;
    my @stops;
    my $r_par1;
    my $r_par2;

    if ($self->flag_session__are_we_in_random_walk_session) {
        ($r_par1, $r_par2) = 
            $self->get_random_walk_range;
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        ($r_par1, $r_par2) = 
            $self->get_svm_range;
    }
    elsif ($self->flag_session__are_we_in_cycle_session) {
        $r_par1 = 
            $self->get_cycle_train_range;
        $r_par2 = (undef,undef,undef); #competability with the previous ones
    }
    else {
        confess "qsub_setting_range error\n";
    }

    push @starts, $r_par1->[0];
    push @starts, $r_par2->[0];

    push @gaps, $r_par1->[1];
    push @gaps, $r_par2->[1];

    push @stops, $r_par1->[2];
    push @stops, $r_par2->[2];


    #again competability with cycle range
    @starts = grep {defined $_} @starts;
    @gaps = grep {defined $_}   @gaps;
    @stops = grep {defined $_}  @stops;


    my ($tot_per, $r_lengths) = $sge->set_range(\@starts,\@stops,\@gaps);
    if ($self->flag_verbose_get) {
        print "Total number of permutation is $tot_per\n",
              "The lengthes of the parameters are:",
              "@$r_lengths \n";
    }
    $self->set_range_in_cluster_units($tot_per);
    $self->set_grid_lengths($r_lengths);
    $self->flag_qsub_range_set;

    return ($tot_per,$r_lengths);
}


sub qsub_script_add_all_files {
    #should be used only after all the parameters were insterted

    my $self = shift;
    my $sge = $self->qsub_script_get;

    $self->test_preparation_all;

    my ($r_referes, $r_fullnames) = $self->get_mandatory_files;


    foreach my $i (0..scalar @$r_fullnames-1) {
        my $filename = $r_fullnames->[$i];
        my $refer    = $r_referes->[$i];

        $sge->add_required_file( {refer=> $refer,
                                  name => $filename,
                                  place=> 'local'});
    }

    $sge->add_required_file ({refer=> 'hash_pos_to_seq_file',
                              name=>   (abs_path ($self->get_file_pos_to_seq)),
                              place=> 'local'});

    $sge->add_required_file ({refer=> 'unify_int_file',
                              name=>   (abs_path ($self->get_unify_int_file)),
                              place=> 'local'});

    $self->{$sge} = $sge;

    return 1;
}

sub qsub_add_command_line {
    my $self = shift;
    confess "qsub_add_command_line error\n"
        if (not $self->flag_command_line_get);

    my $command_line = $self->get_command_line;
    my $sge = $self->qsub_script_get;

    $sge->add_command ($command_line);
    $self->{$sge} = $sge;
    return 1;
}

sub qsub_compile_and_test {
    my $self = shift;
    my $sge = $self->qsub_script_get;
    my $filename = $sge->compile;
    $sge->check_integrity;
    $self->set_sge_script($filename);

    return $filename;
}


sub set_sge_script {
    my ($self,$file) = @_;
    confess "set_sge_script error\n"
        if (not defined $file);
    $self->{sge_perl_script} = $file;
    return $self->{sge_perl_script};
}

sub get_sge_script {
    my $self = shift;
    return $self->{sge_perl_script};
}

sub generate_command_line {

    my $self = shift;

    my $command_line;
    my $random_walk;
    my $svm_params;
    my $train_cycles;


    my $unify_int_file          = '<unify_int_file>';
    my $hash_pos_to_seq_file    = '<hash_pos_to_seq_file>';
    my $number_of_cycles        = $self->get_number_of_cycles;
    my $counter_threshold       = $self->get_counter_threshold;
    my $cross_vald              = $self->get_cross_val;
    my $estimate_prop           = $self->get_estimate_prop;
    my $bring_files             = $self->get_bring_back_files_to;


    if ($self->flag_session__are_we_in_random_walk_session) {
        $random_walk            = _wild_card();
        $svm_params             = join $SPACE, $self->get_c_g;
        $train_cycles           = join $SPACE, $self->get_train_cycles;
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        $random_walk            = join $SPACE, $self->get_p1_p2;
        $svm_params             = _wild_card();
        $train_cycles           = join $SPACE, $self->get_train_cycles;
    }
    elsif ($self->flag_session__are_we_in_cycle_session) {
        $random_walk            = join $SPACE, $self->get_p1_p2;
        $svm_params             = join $SPACE, $self->get_c_g;
        $train_cycles           = _wild_card();
    }
    else {
        confess 
            "generate_command_line error:", 
            "non of the session flags is set\n";
    }

    $command_line = join ($SPACE,
                        $REMOTE_SCRIPT_NANE      ,
                        '--random_walk'          ,  $random_walk            ,
                        '--svm'                  ,  $svm_params             ,
                        '--hash_pos_to_seq_file' ,  $hash_pos_to_seq_file   ,
                        '--number_of_cycles'     ,  $number_of_cycles       ,
                        '--unify_int_file'       ,  $unify_int_file         ,
                        '--counter_threshold'    ,  $counter_threshold      ,
                        '--train_cycles'         ,  $train_cycles           );

    if ($cross_vald) {
        $command_line .=    $SPACE. '--cross_vald'. $SPACE . $cross_vald;
    }

    if ($estimate_prop) {
        $command_line .=    $SPACE. '--prob_estimate';
    }

    if ($bring_files) {
        $command_line .=    $SPACE. '--bring_files' . $SPACE . $bring_files;
    }

    $command_line .= $SPACE;

    $self->{command_line} = $command_line;
    $self->flag_command_line_set;
    return $command_line;

}

sub get_command_line {
    my $self = shift;
    return $self->{command_line};
}

######################################### submit! #########################

sub submit_quicky {
    my $self = shift;

    my $update_time = shift @_;
    confess "submit_automatic_pilot error - no update time\n"
        if (not $update_time);

    my $perl_file_to_submit = $self->get_sge_script
        or confess "no perl file to submit\n";

    if (not -e $perl_file_to_submit) {
        confess "$perl_file_to_submit is not found\n";
    }

    my $life_time = $self->submit_life_time_get or 
        confess "life time is requiered\n";

    my $cluster_units
             = $self->get_range_in_cluster_units or
                confess "range is requiered\n";

    if ($self->flag_verbose_get) {
        print "Submitted new job on:",time()            ,"\n",
              "Life time:\t"        ,$life_time,"Hours" ,"\n",
              "Number of Tasks:\t"  ,$cluster_units     ,"\n",
              "Update rate:\t"      ,$update_time,"Mins","\n";
    }

    my $submiter = $self->submit_new;
    $submiter->life_time($life_time);
    $submiter->range(1,$cluster_units,1);
    $submiter->submit($perl_file_to_submit);
    if ($self->flag_session__are_we_in_cycle_session) {
        $submiter->memory($MEMORY_FOR_CYCLE_SESSION);
        if ($self->flag_debug_mode_get) {
            print "Memory:", $submiter->memory,"\n";
        }
    }
    $submiter->automatic_pilot($update_time);

    $self->flag_finish_submitter_set;

    return 1;
}

sub submit_new {

    my $self = shift;
    my $submiter = Submit_tools->new();
    $self->{submit}->{object} = $submiter;
    return $self->{submit}->{object};
}

sub submit_update_object {
    my $self = shift;
    my $submiter = shift;
    confess "submit_update error\n"
        if (not $submiter);
    $self->{submit}->{object} = $submiter;
    return $self->{submit}->{object};
}

sub submit_get_object {
    my $self = shift;
    return $self->{submit}->{object};
}

sub submit_life_time_set {
    my $self = shift;
    my $life_time = shift @_;

    confess "submit_life_time_set error\n"
            if (not $life_time);
    $self->{submit}->{life_time} = $life_time;
    return $self->{submit}->{life_time};
}

sub submit_life_time_get {
    my $self = shift;
    return $self->{submit}->{life_time};
}

sub set_range_in_cluster_units {
    my $self = shift;
    my $cluster_units = shift @_;
    confess "set_range_in_cluster_units error\n"
        if (not $cluster_units);

    $self->{submit}->{cluster_units} = $cluster_units;
    return $self->{submit}->{cluster_units};
}

sub get_range_in_cluster_units {
    my $self = shift;
    return $self->{submit}->{cluster_units};
}
sub set_grid_lengths {
    my $self = shift;
    my $r_lengths = shift;
    confess "set_grid_lengths error ",
          "input value should be array\n"
            if (ref $r_lengths ne 'ARRAY');

    $self->{grid_lengths} = $r_lengths;
    return $r_lengths;
}

sub get_grid_lengths {
    #return reference
    my $self = shift;
    return 
        $self->{grid_lengths};
}


######################################### Communicator #########################
sub stdout_parse_matrix_quickly {
    my $self = shift;

    confess "stdout_parse_matrix_quickly can be activated only when submitter is finished\n"
        if (not $self->flag_finish_submitter_get);

    my @stdout_files = ($self->submit_get_object)->get_message_filenames;

    $self->stdout_files_set(@stdout_files);
    $self->test_stdout_do_we_have_all_files;
    $self->test_location_stdout_files;
    $self->stdout_construct_matrices($self->get_train_cycles);
    $self->stdout_parse_all_files____for_matrix;
    $self->stdout_clean;
    return 1;

}


sub stdout_parse_returned_files_quickly {
    my $self = shift;
    confess "stdout_parse_returned_files_quickly can be activated only when submitter is finished\n"
        if (not $self->flag_finish_submitter_get);
    my @stdout_files = ($self->submit_get_object)->get_message_filenames;

    $self->stdout_files_set(@stdout_files);
    $self->test_stdout_do_we_have_all_files;
    $self->test_location_stdout_files;
    $self->stdout_parse_all_files____for_returened_files;
    $self->stdout_clean;

}




sub stdout_clean {

    my $self = shift;
    my @files = $self->stdout_files_get;
    @files = map {abs_path $_ } @files;

    if ($self->flag_verbose_get) {
        print "Cleaning stdout files\n";
    }

    my $job_name = ($self->submit_get_object)->get_job_id ;

    #if debug mode keep the files:
    if ($self->flag_debug_mode_get) {
        my $garb_dir_fullname = _make_grabage_directory($job_name);
        my $tar = Archive::Tar->new();
        $tar->add_files(@files);
        $tar->write("$garb_dir_fullname/$job_name.tar.gz",1);
        print "files were tarred in $garb_dir_fullname under $job_name.tar";
    }

    foreach my $file (@files) {
        unlink $file ||
            confess "can't clean file: $file $!\n";
        if ($self->flag_verbose_get) {print "$file cleaned\n"};
    }
}


sub stdout_files_set {
    my $self = shift;
    if (@_) {
        @{$self->{communicator}->{files}} = @_;
        return @{$self->{communicator}->{files}};
    }
    confess "stdout_files_set error\n";
    return 0;
}

sub stdout_files_get {
    my $self = shift;
    return @{$self->{communicator}->{files}};
}

sub stdout_parse_all_files____for_matrix {
    my $self = shift;

    my @train_cycles = $self->get_train_cycles;

    FILE:
     foreach my $file ($self->stdout_files_get) {
        my ($r_parameters,$r_result_of) = 
            $self->stdout_parse_file___for_matrix($file);

        my ($par1,$par2) = @$r_parameters;

        my ($d1,$d2) = 
            $self->stdout_convert_param_to_cord($par1,$par2);

        $self->stdout_verify_cord($d1,$d2);

        TRAINING_CYCLE:
        foreach my $cycle (@train_cycles) {
            my $result = $r_result_of->{$cycle};
            if (not defined $result) {
                $result = $D_UNDEFINED_VALUE_IN_MATRIX;
                carp "could not find result for $par1,$par2,$cycle\n",
                     "maybe timeout of submittion?\n",
                     "It is ok, I will continue processing\n";
            }
            my $matrix = $self->stdout_matrix_get($cycle);
            set($matrix,$d1,$d2,$result);
            $self->stdout_matrix_update($matrix,$cycle);
        }
    }
    return 1;
}
sub stdout_parse_all_files____for_returened_files {
    my $self = shift;
    my %lookup_table_for_all_files;

    my $count = 0;
    foreach my $file ($self->stdout_files_get) {
        $count++;

        my $r_lookup_table_for_file = 
            $self->stdout_parse_file___for_returned_files($file);

        my $filename = $r_lookup_table_for_file->{Tar_at_home};
        $self->test_returen_tar($filename,$r_lookup_table_for_file);
        $lookup_table_for_all_files{$count} = $r_lookup_table_for_file;
    }

    if ($self->flag_debug_mode_set) {
        print "This is the table that stored in the hash:\n";
        print Dumper (%lookup_table_for_all_files);
        print "\n\n\n";
    }


    $self->stdout_looktable_for_all_files_set(\%lookup_table_for_all_files);
    return $count;
}

sub stdout_looktable_for_all_files_set {
    my $self = shift;
    my $r_lookup_table_for_file = shift;
    confess "must give hash reference to stdout_looktable_for_all_files_set\n"
        if (ref $r_lookup_table_for_file ne 'HASH');
    $self->{stdout}->{lookup_table_for_all_files} = $r_lookup_table_for_file;
}

sub stdout_looktable_for_all_files_get {
    my $self = shift;
    return 
    $self->{stdout}->{lookup_table_for_all_files};
}

sub stdout_looktable_for_all_files__give_number_of_files {
    my $self = shift;
    return scalar (keys %{$self->stdout_looktable_for_all_files_get});
}
sub stdout_looktable__search {
    my $self = shift;
    my $r_hash_to_match = shift;
    confess "must give hash reference to stdout_looktable_search\n"
        if (ref $r_hash_to_match ne 'HASH');

    my $n = $self->stdout_looktable_for_all_files__give_number_of_files;
    my $r_table = $self->stdout_looktable_for_all_files_get;


    ENTRY:
    foreach my $entry (1..$n) {

        my $correct_answers = 0;#how many corrcet answer we got?

        FIELDS:
        while (my ($k_search,$v_search) = each (%$r_hash_to_match)) {
            if ($r_table->{$entry}->{$k_search} eq $v_search) {
                $correct_answers++;
            }
        } #end of FIELDS (iteration over all the conditions of the query
        if ($correct_answers == scalar keys %$r_hash_to_match) {
            my $filename = $r_table->{$entry}->{'Tar_at_home'};
            return $filename;
        }
    } #end of ENTRY (iteration over all entries)
    return;
}



sub stdout_parse_file___for_returned_files {
    my $self = shift;
    my $file = shift || 
        confess "stdout_parse_file___for_returned_files error",
              " - no file was entered\n";

    my %lookup_table_for_file;

    open my $FH,'<',$file
        or confess "can't open $file : $!\n";

    while (<$FH>) {
        chomp;
        if (m/^>(\w+):(.+)/) {
            $lookup_table_for_file{$1} = $2;
        }
    }
    close $FH;
    return \%lookup_table_for_file;
}

sub stdout_construct_matrices {
    my $self = shift;
    my @cycles = @_;
    confess "stdout_construct_matrices error\n"
        if (scalar @cycles < 1);

    confess "for strictness - matrices can be constucted ",
          "only after setting qsub ranges\n"
            if (not $self->flag_qsub_range_get);

    my $r_grid_lengths = $self->get_grid_lengths;

    foreach my $cycle (@cycles) {
        my $pdl = zeroes @$r_grid_lengths;
        $self->{communicator}->{matrix}->{$cycle} = $pdl;
    }

    my $pdl = zeroes @$r_grid_lengths;
    $self->{communicator}->{__matrix}->{proto} = $pdl;
        #proto matrix is not part of the matrix hash!
    return 1;
}


sub stdout_matrix_update {
    my $self    =   shift;
    my $matrix  =   shift;
    confess "stdout_matrix_update error\n"
        if (ref $matrix ne 'PDL');

    my $cycle   =   shift ||
                confess "stdout_matrix_update error",
                "must enter cycle\n";

    $self->{communicator}->{matrix}->{$cycle} = $matrix;
    return $self->{communicator}->{matrix}->{$cycle};
}

sub stdout_matrix_get {
    my $self = shift;
    my $cycle = shift ||
        confess "stdout_matrix_get error - must enter cycle\n";
    return $self->{communicator}->{matrix}->{$cycle};
}

sub stdout_matrix_prototype_get {
    my $self = shift;
    return $self->{communicator}->{__matrix}->{proto};
        #proto matrix is not part of the matrix hash!
}

sub stdout_number_of_matrices {
    my $self = shift;
    return scalar keys %{$self->{communicator}->{matrix}};
}

sub stdout_matrix_get_cycles_with_matrix {
    my $self = shift;
    return sort {$a <=> $b} 
                keys %{$self->{communicator}->{matrix}};
}


sub stdout_convert_param_to_cord {
    my $self = shift;
    confess "stdout_convert_param_to_cord ",
          "need 2 parameters\n"
                if (scalar @_ <2);

    my ($d1,$d2);
    if ($self->flag_session__are_we_in_random_walk_session) {
        ($d1,$d2) = 
            $self->_stdout_convert_param_to_cord_r(@_);
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        ($d1,$d2) = 
            $self->_stdout_convert_param_to_cord_s(@_);
    }
    else {
        confess "can't find session type\n";
    }

    $self->stdout_verify_cord($d1,$d2);

    return ($d1,$d2);
}

sub _stdout_convert_param_to_cord_r {
    my $self = shift;
    my ($p1,$p2) = @_;
    my ($r_p1,$r_p2) = 
        $self->get_random_walk_range;
    my ($p1_s,$p1_g,undef) = @$r_p1;
    my ($p2_s,$p2_g,undef) = @$r_p2;
    my $d1 = ($p1 - $p1_s) / $p1_g;
    my $d2 = ($p2 - $p2_s) / $p2_g;

    $d1 = round ($d1);
    $d2 = round ($d2);

    return ($d1,$d2);
}

sub _stdout_convert_param_to_cord_s {
    my $self = shift;
    my ($c,$gamma) = @_;
    my ($r_c,$r_gamma) = 
        $self->get_svm_range;
    my ($c_s,$c_g,undef) = @$r_c;
    my ($gamma_s,$gamma_g,undef) = @$r_gamma;
    my $d1 = ($c - $c_s) / $c_g;
    my $d2 = ($gamma - $gamma_s) / $gamma_g;
    return ($d1,$d2);
}


sub stdout_verify_cord {
    my ($self,$d1,$d2) = @_;

    my $matrix = $self->stdout_matrix_prototype_get;
    confess "matrix is not defined\n"
        if (ref $matrix ne 'PDL');

    my ($d1_size,$d2_size) = dims $matrix;


    confess "stdout_verify_cord error: ",
          "coordinates should be intergers: ",
          "$d1,$d2\n"
            if (
                (abs (round ($d1) - $d1) > $EPSILON) or 
                (abs (round ($d2) - $d2) > $EPSILON) 
               );



    confess "stdout_verify_cord error: ",
          "corrdiantes are outside the matrix: ",
          "$d1,$d2 and matrix size is: ",dims $matrix,
          "\n"
            if (
                ($d1 > $d1_size)        or
                ($d2 > $d2_size)        or
                ($d1 < 0)               or
                ($d2 < 0) 
               );

    return 1;
}



sub stdout_parse_file___for_matrix {

    my $self = shift;
    my $file = shift || 
        confess "stdout_parse_file___for_matrix error",
              " - no file was entered\n";

    my %params;
    my $result = undef;
    my @params_a;

    my %results_of_cycle;
    my @cycles_allowed = $self->get_train_cycles;
    lock_keys (%results_of_cycle,@cycles_allowed);


    if ($self->flag_session__are_we_in_random_walk_session) {
        %params = ('p1' => undef,
                   'p2' => undef,
                  )
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        %params = ('c' => undef,
                   'gamma' => undef,
                  )
    }
    else {
        confess "session type is not supported by",
              " stdout_parse_file___for_matrix\n";
    }

    open my $FH, '<', $file 
        or confess "can't open file to parse: $!\n";


    my $cycle=undef;
    my $line_count=0;
    while (<$FH>)   {
        $line_count++;
        chomp;
            if (
                (m/^>(\w+):(.+)/)    and
                (exists $params{$1})
                                    ) {#>p1:0.12
                $params{$1} = $2;
            }
            elsif (m/>cycle:(\d+)</) {
                $cycle = $1;
            }
            elsif (m/^Cross Validation Accuracy\s=\s(.*)%/) {
                $result = $1;
                $results_of_cycle{$cycle} = $result;
            }
    }

    close $FH;


    foreach my $key (sort keys %params) {
        #super dengerous code - the order
        #of the reported values is according
        #to the keys alphabeta!

        my $grid_v = $params{$key};
        confess "can't find the value of $key\n"
            if (not defined $grid_v);

        push @params_a,$grid_v;
    }


    if ($self->flag_verbose_get) {
        print "Parsing file: $file for cross validation\t",
              "found $line_count lines in the file\n",
              "parameters found: @params_a\n", "Values: ", 
               scalar grep {defined $_} keys %results_of_cycle;

        print " values, out of possible ",scalar @cycles_allowed,"\n","\n";
    }


    return (\@params_a,\%results_of_cycle);
}
######################################### Matrix Analysis #########################
sub matanl_maximum_upon_average {

    my $self = shift;
    my  @cycles = $self->get_train_cycles;
    confess "no cycles to process\n"
        if (not scalar @cycles);


    my $matrix_average = $self->stdout_matrix_prototype_get;
    foreach my $cycle (@cycles) {
        my $matrix = $self->stdout_matrix_get($cycle);
        $matrix_average += $matrix;
    }

    $matrix_average = $matrix_average / (scalar (@cycles));

    my ($par_1_pos_max,$par_2_pos_max,$max) = 
        $self->matanl_find_maximum_point($matrix_average);

   if ($self->flag_verbose_get) {
        print "Finding maximum values across all cycles..",
        " On average: $par_1_pos_max and $par_2_pos_max are the best",
        " with $max% cross validation rate\n";
    }

    if ($self->flag_debug_mode_set) {
        print "The average matrix:\n";
        print $matrix_average;
    }

    return ($par_1_pos_max,$par_2_pos_max,$max);
}


sub matanl_find_maximum_point {
    my $self = shift;
    my $matrix = shift;
    confess "matanl_find_maximum_point error - matrix is not defined\n"
        if (ref $matrix ne 'PDL');

    my $max = max $matrix;

    my ($corr_1_size,$corr_2_size) = dims $matrix;
    my ($corr_1_pos_max,$corr_2_pos_max);

    foreach my $corr_1_pos (0..$corr_1_size-1) {
        foreach my $corr_2_pos (0..$corr_2_size-1) {
            my $v = at($matrix,$corr_1_pos,$corr_2_pos);
            ($corr_1_pos_max,$corr_2_pos_max) = ($corr_1_pos,$corr_2_pos)
                if ($v == $max);
        }
    }

    my ($par_1_pos_max,$par_2_pos_max) = 
            $self->matanl_convert_corr_to_par($corr_1_pos_max,$corr_2_pos_max);

    return ($par_1_pos_max,$par_2_pos_max,$max);
}






sub matanl_convert_corr_to_par {
    my $self = shift;
    my ($d1,$d2) = @_;
    my ($par1,$par2);

    if ((not defined $d1) or 
         (not defined $d2)) {
        confess "matanl_convert_corr_to_par error\n";
    }

    if ($self->flag_session__are_we_in_random_walk_session) {
        my ($r_p1,$r_p2) = 
                $self->get_random_walk_range;
        my ($p1_s,$p1_g,$p1_e) = @$r_p1;
        my ($p2_s,$p2_g,$p2_e) = @$r_p2;

        $par1 = ($d1 * $p1_g) + $p1_s;
        $par2 = ($d2 * $p2_g) + $p2_s;

    }
    elsif ($self->flag_session__are_we_in_svm_session) {

        my ($r_c,$r_gamma) = 
            $self->get_svm_range;
        my ($c_s,$c_g,$c_e) = @$r_c;
        my ($gamma_s,$gamma_g,$gamma_end) = @$r_gamma;

        $par1 = ($d1 * $c_g) + $c_s;
        $par2 = ($d2 * $gamma_g) + $gamma_s;

    }
    else {
        confess "matanl_convert_corr_to_par is only",
              "for SVM or random walk session\n";
    }
    return ($par1,$par2);
}



sub matanl_verifty_parr {
    my $self = shift;
    if (scalar @_ < 2) {
        confess "matanl_verifty_parr error\n";
    }

    my ($par1,$par2) = @_;
    my ($r_range_par1,$r_range_par2);


    if ($self->flag_session__are_we_in_random_walk_session) {
        ($r_range_par1,$r_range_par2) =  $self->get_random_walk_range;
    }
    elsif ($self->flag_session__are_we_in_svm_session) {
        ($r_range_par1,$r_range_par2) =  $self->get_svm_range;
    }
    else {
        confess "matanl_verifty_parr is only","for SVM or random walk session\n";
    }

    my ($start,undef,$end) = @$r_range_par1;
    if(not _am_I_there($par1,$start,$end)) {
        confess "matanl_verifty_parr error: first parameter out of range: $par1",
              " (range is between $start to $end)\n"
    }
    ($start,undef,$end) = @$r_range_par2;
    if(not _am_I_there($par2,$start,$end)) {
        confess "matanl_verifty_parr error: second parameter out of range: $par2",
              " (range is between $start to $end)\n"
    }
    return 1;
}


######################################### Aux #########################
sub maxI {
    my @q = sort {$a <=> $b} @_;
    my $w = pop @q;
    return $w;
}

sub _wild_card {
    my $wildcard = '<X>';
    return $wildcard;
}

sub _make_grabage_directory {
    my ($subdir) = @_;
    if (not -d './tmp') {
        mkdir 'tmp'
            or confess "can't make tmp dir:$!\n";
    }
    my $subdir_c = $subdir;
    while (-d "./tmp/$subdir_c") {
        $subdir_c = $subdir . time();
        #carp "chang name of tmp subdir to $subdir_c\n";
    }
    mkdir "./tmp/$subdir_c"
        or confess "can't make tmp subdir $subdir_c: $!\n";

    $subdir_c = abs_path "./tmp/$subdir_c";
    return $subdir_c;
}

sub _am_I_there {
    my ($parameter,$start,$end) = @_;
    if (($parameter >= $start) and 
       ($parameter <= $end))      {

        return 1;
    }
    return 0;
}

sub round {
    my ($number) = @_;
    return int($number + .5 * ($number <=> 0));
}
1;
