#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)



package Model;
use strict;
use warnings;
use Carp;
use Readonly;
use Time::HiRes qw(gettimeofday);
use PDL;
use Data::Dumper;

sub set_parameters {
	my $self = shift;
	if (@_) {@{$self->{'parameters'}} = @_}
}
sub get_parameters {
	my $self = shift;
	return @{$self->{'parameters'}};
}

sub end_cycle {
	my $self = shift;
	if (@_) {$self->{'CYCLE_END'} = shift};
	return $self->{'CYCLE_END'};
}

sub set_sequence {
	my $self = shift;
	if (@_) {
		my $seq = shift;
		$seq =~ tr/agct/AGCT/;
		$self->{'sequence'} = $seq;
		my @t = split // , $seq;
		$self->{'_sequence_length'}=scalar (@t);
                my $pdl_sequence = _convert_sequence_to_matrix($seq,scalar(@t));
                $self->{'sequence_matrix'} = $pdl_sequence;
	};
}

sub get_sequence {
	my $self = shift;
	return $self->{'sequence'};
}

sub get_sequence_matrix {
    my $self = shift;
    return $self->{'sequence_matrix'};
}

sub print_sequence_matrix {
    my $self = shift;
    print $self->{'sequence_matrix'};
}



sub _convert_sequence_to_matrix {

    my ($seq,$length) = @_;
    my @mat_as_array;

    if (not defined $length) { $length = length ($seq)};

    $seq = substr ($seq,0,$length);
    $seq=uc($seq);

    my $seqCopy = $seq;
    $seqCopy =~tr/ACGT/1000/;
    @{$mat_as_array[0]} = split //, $seqCopy;

    $seqCopy = $seq;
    $seqCopy =~tr/ACGT/0100/;
    @{$mat_as_array[1]} = split //, $seqCopy;

    $seqCopy = $seq;
    $seqCopy =~tr/ACGT/0010/;
    @{$mat_as_array[2]} = split //, $seqCopy;

    $seqCopy = $seq;
    $seqCopy =~tr/ACGT/0001/;
    @{$mat_as_array[3]} = split //, $seqCopy;

    my $matrix = pdl (\@mat_as_array);
    $matrix = $matrix->transpose;
    return $matrix;
}



1;
